/*
 * GEOSYNC.C - Calculate position of a geostationary satellite.
 * Version 1.2:	S.Hosgood, May 10th 1989
 * Version 2.0: S.Hosgood, Jan 18th 1996
 *
 * This program Copyright (c), Steve Hosgood 1996.
 * It may be freely copied and distributed as long as no charge is made.
 *
 * No responsibility can be accepted for the accuracy of the results.
 * The program is for amusement only.
 *
 * Written for the IBM PC, using Microsoft C 5.1
 * Use is made of the IBM's 'degrees' symbol.
 * To run: just type 'geosync' and answer the questions.
 *
 * New in V2.0: if the environment variables ICBM_LAT and/or ICBM_LONG are
 * set then they will normally be used, and the user will not be prompted
 * for these values. The command-line flag "-i" overrides this, and the
 * values will always be prompted-for.
 *
 * Please forward bug reports to:
 * Steve Hosgood (iisteve@iiit.swan.ac.uk).
 *
 */

# include	<stdio.h>
# include	<stdlib.h>
# include	<math.h>
# include	<ctype.h>

# ifndef	PI
# define	PI		3.141592654
# endif

# define	degtorad(x)	((x)*PI/180.0)
# define	radtodeg(x)	((x)*180.0/PI)

/* Radius of Earth, and Radius of Geosync Orbit */
/* These constants are off the top of my head, may be a bit wrong */
# define	E_RAD		6378.099	/* Km (equator) */
# define	P_DIFF		21.468		/* difference at poles */
# define	O_RAD		42240.31	/* Km */

double readval( char *prompt, char *env_var );

int
main( int argc, char *argv[] )
{
	int	iflag = 0;
	double	a, b, c, d, n, w;
	double	vert, tmp, e_rad;
	double	satwest, obswest, obsnorth;

	if (strcmp(argv[1], "-i") == 0)
		iflag = 1;

	satwest = readval("Enter satellite's Longitude", NULL);
	obswest = readval("Enter observer's Longitude", iflag? NULL: "ICBM_LONG");
	obsnorth = readval("Enter observer's Latitude", iflag? NULL: "ICBM_LAT");

	a = sin(w = degtorad(satwest - obswest));
	b = sin(n = degtorad(obsnorth));
	c = cos(w);
	d = cos(n);

	/* Try to allow for local earth radius being dependant on latitude */
	/* I'm not sure of the shape factor here */
	/* Assume radius follows a cos-law with latitude */
	e_rad = E_RAD - P_DIFF*(1 - c);
	vert = O_RAD*c*d - e_rad;

	/* attempt to avoid DOMAIN errors from trig functions */
	if ((tmp = O_RAD*sqrt(c*c*b*b + a*a)) <= 1e-6 && tmp >= -1e-6) {
		if (vert > 0.0)
			printf("Vertically up\n");
		else
			printf("Vertically down\n");
	}
	else {
		printf("Elevation = %6.2f\370\n", radtodeg(atan2(vert, tmp)));

		if ((tmp = c*b) <= 1e-6 && tmp >= -1e-6)
			/* bearing directly east or west */
			tmp = a > 0.0? 90.0: -90.0;
		else
			tmp = radtodeg(atan2(a, tmp));


		/* nice to see bearings +/- South or +/- North */
		printf("Bearing (from ");

		if (tmp < -90.0 || tmp > 90.0) {
			tmp += 180.0;
			printf("north");

			if (tmp >= 180.0)
				tmp -= 360.0;
		}
		else
			printf("south");

		printf(") = %6.2f\370\n", tmp);
	}

	if ((d = e_rad/(O_RAD*cos(n))) > 1.0 || d < -1.0) {
		printf("No geosync satellites visible from there\n");
		printf("Limit latitude is appx \361%6.2f\370\n", radtodeg(acos(E_RAD/O_RAD)));
	}
	else {
		c = radtodeg(acos(d));

		fputs("Satellites ", stdout);
		prlong(c+obswest);
		fputs(" thru ", stdout);
		prlong(obswest-c);
		fputs(" are visible\n", stdout);
	}

	exit(0);
}

double
readval( char *str, char *env_var )
{
	char		buffer[30];
	register char	chr, *cptr = NULL;
	int		err;
	double		ret;

	if (env_var)
		cptr = getenv(env_var);

	if (cptr)
		strncpy(buffer, cptr, sizeof buffer);
	else {
		/* prompt */
		fprintf(stderr, "%s: ", str);
		cptr = fgets(buffer, sizeof buffer, stdin);
	}

	do {
		err = 0;

		if (*cptr == '-')
			cptr++;

		/* get integer degrees part anyway */
		ret = atoi(cptr);

		/* look for next non-numeric */
		for (; (chr = *cptr) && isdigit(chr); cptr++);

		if (chr == ':') {
			/* get integer of minutes of arc */
			ret += atoi(++cptr) / 60.0;

			/* step past minutes of arc */
			for (; (chr = *cptr) && isdigit(chr); cptr++);

			if (chr == ':') {
				/* get seconds of arc */
				ret += atof(++cptr) / 3600.0;

				/* step past seconds of arc */
				for (; (chr = *cptr) && isdigit(chr); cptr++);
				if (chr == '.')
					/* step past optional decimal part */
					for (cptr++; (chr = *cptr) && isdigit(chr); cptr++);
			}
			else if (chr == '.') {
				/* get decimal part of minutes of arc */
				ret += atof(cptr) / 60.0;

				/* step past decimal part */
				for (cptr++; (chr = *cptr) && isdigit(chr); cptr++);
			}

		}
		else if (chr == '.') {
			/* get decimal part of degrees */
			ret += atof(cptr);

			/* step past decimal part */
			for (cptr++; (chr = *cptr) && isdigit(chr); cptr++);
		}

		/* check final character */
		switch (chr) {
		case 'N':
		case 'n':
		case 'W':
		case 'w':
		case '\n':
		case '\0':
			/* 'North' or 'West' assumed - do nothing */
			break;

		case 'E':
		case 'e':
		case 'S':
		case 's':
			ret = -ret;
			break;

		default:
			/* print a fairly useful error message */
			fputs(buffer, stderr);
			*(cptr+1) = '\0';
			*cptr-- = '^';
			while (cptr >= buffer)
				*cptr-- = ' ';

			fprintf(stderr, "%s-----Error in input\n\n", buffer);
			err = 1;
		}

		if (err) {
			fprintf(stderr, "Format: {-}<deg>{NSEW}\n");
			fprintf(stderr, "<deg> ::= integer or real\n\n");
			fprintf(stderr, "Or..  : {-}<deg>:<min>{NSEW}\n");
			fprintf(stderr, "<deg> ::= integer\n");
			fprintf(stderr, "<min> ::= integer or real\n\n");
			fprintf(stderr, "Or..  : {-}<deg>:<min>:<sec>{NSEW}\n");
			fprintf(stderr, "<deg> ::= integer\n");
			fprintf(stderr, "<min> ::= integer\n");
			fprintf(stderr, "<sec> ::= integer or real\n\n");
			fprintf(stderr, "Items in braces are optional\n\n");

			/* re-prompt */
			fprintf(stderr, "%s: ", str);
			cptr = fgets(buffer, sizeof buffer, stdin);
		}
	} while (err);

	return (*buffer == '-')? -ret: ret;
}

prlong(val)
double val;
{
	char	eastwest;

	/* get into range +180 -> -180 degrees */
	while (val > 180.0)
		val -= 360.0;

	while (val < -180.0)
		val += 360.0;

	if (val < 0.0) {
		eastwest = 'E';
		val = -val;
	}
	else if (val > 0.0)
		eastwest = 'W';
	else
		eastwest = ' ';

	printf("%5.1f\370%c", val, eastwest);
	return;
}
